module ELPO where

import Data.List
import Term
import Rule
import TRS
import SMT
import Algebra

--  import Parser

type Parameter = (Precedence, Algebra)

showParameter :: Parameter -> String
showParameter (p, w) =
  "ELPO with interpretations on natural numbers\n\n" ++
  showInterpretations w ++
  "\nand precedence:\n\n" ++
  showPrecedence p ++ "\n"

-- LPO-based SPO

gtLPO :: Parameter -> Term -> Term -> Bool
gtLPO param s t = gtLPO12 param s t || gtLPO3 param s t

gtLPO12 :: Parameter -> Term -> Term -> Bool
gtLPO12 param@(p, a) s@(F _ ss) t@(F _ ts) =
  (gtB p a s t ||
   (geqB a s t && gtLex (gtLPO param) ss ts)) &&
  all (gtLPO param s) ts
gtLPO12 _ _ _  = False
  
gtLPO3 :: Parameter -> Term -> Term -> Bool
gtLPO3 param (F _ ss) t = any (\s -> geqLPO param s t) ss
gtLPO3 _ _ _            = False

geqLPO :: Parameter -> Term -> Term -> Bool
geqLPO param s t = s == t || gtLPO param s t

-- LPO-based MSPO

gt :: Parameter -> Term -> Term -> Bool
gt param@(_, w) s t = geqA w s t && gtLPO param s t


varGtLPO :: TRS -> Term -> Term -> Formula
varGtLPO a s t
  | Just n <- elemIndex (s, t) a = FVar ("varGtLPO" ++ show n)
  | otherwise                    = error "varGtLPO"

encodeGeqLPO :: TRS -> Term -> Term -> Formula
encodeGeqLPO a s t
  | s == t    = top
  | otherwise = varGtLPO a s t

encodeGtLPO :: TRS -> Term -> Term -> Formula
encodeGtLPO a s@(F _ _) t = disj [encodeGtLPO1 a s t, encodeGtLPO23 a s t]
encodeGtLPO _ _         _ = bottom

encodeGtLPO1 :: TRS -> Term -> Term -> Formula
encodeGtLPO1 a (F _ ss) t = disj [ encodeGeqLPO a s t | s <- ss ]
encodeGtLPO1 _ _        _ = bottom

encodeGtLPO23 :: TRS -> Term -> Term -> Formula
encodeGtLPO23 a s@(F _ ss) t@(F _ ts) =
  disj [conj (encodeGtB s t : [ varGtLPO a s tj | tj <- ts ]),
        conj [encodeGeqB s t, encodeLexLPO a s ss ts]]
encodeGtLPO23 _ _ _ = bottom

encodeLexLPO :: TRS -> Term -> [Term] -> [Term] -> Formula
encodeLexLPO _ _ [] [] = Or []
encodeLexLPO a u (s : ss) (t : ts)
  | s == t    = encodeLexLPO a u ss ts
  | otherwise = conj (varGtLPO a s t : [ varGtLPO a u v | v <- ts ])
encodeLexLPO _ _ _ _ = error "encodeLexSPO"

varGtELPO :: TRS -> Term -> Term -> Formula
varGtELPO a s t
  | Just n <- elemIndex (s, t) a = FVar ("gtELPO" ++ show n)
  | otherwise                    = error "varGtELPO"

encodeGtELPO :: TRS -> Term -> Term -> Formula
encodeGtELPO a s t = conj [encodeGeqA s t, varGtLPO a s t]

encode :: Signature -> TRS -> (Formula, [(Rule, Formula)])
encode sig trs =
  (conj (Distinct [ varPrec f | (f, _) <- sig ] :
         [ Geq (varA f) (Val 0) | (f, _) <- sig ] ++
         [ Geq (varA (mark f)) (Val 0) | (f, _) <- sig ] ++
         [ iff (varGtLPO a s t) (encodeGtLPO a s t) | (s, t) <- a ] ++
         [ iff (varGtELPO a s t) (encodeGtELPO a s t) | (s, t) <- trs ]),
   [ ((l, r), varGtELPO a l r) | (l, r) <- trs ])
  where 
    a = nub [ (s, t) | (l, r) <- trs, s <- subterms l, t <- subterms r ]

-- Decoding functions.

decode :: Signature -> Model -> Parameter
decode sig model =
  (decodePrecedence model, decodeAlgebra model sig)
